

/*


lzsa_dst+1 - where unpack
lzsa_src+1 - from

//-----------------------------------------------------------------------------
//decompress raw lzsa1 block. create one with lzsa -r <original_file> <compressed_file>
;
//in:
//* lzsa_src_lo and lzsa_src_hi contain the compressed raw block address
//* lzsa_dst_lo and lzsa_dst_hi contain the destination buffer address
;
//out:
//* lzsa_dst_lo and lzsa_dst_hi contain the last decompressed byte address, +1
;
//-----------------------------------------------------------------------------
//backward decompression is also supported, use lzsa -r -b <original_file> <compressed_file>
//to use it, also define backward_decompress=1 before including this code!
;
//in:
//* lzsa_src_lo/lzsa_src_hi must contain the address of the last byte of compressed data
//* lzsa_dst_lo/lzsa_dst_hi must contain the address of the last byte of the destination buffer
;
//out:
//* lzsa_dst_lo/lzsa_dst_hi contain the last decompressed byte address, -1
;
//-----------------------------------------------------------------------------
;
// copyright (c) 2019 emmanuel marty, peter ferrie
;
// this software is provided 'as-is', without any express or implied
// warranty.  in no event will the authors be held liable for any damages
// arising from the use of this software.
;
// permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
;
// 1. the origin of this software must not be misrepresented//you must not
//    claim that you wrote the original software. if you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. this notice may not be removed or altered from any source distribution.
//-----------------------------------------------------------------------------
*/

decompress_lzsa1_fast:
   ldy #$00

decode_token:
   jsr getsrc                           // read token byte: o|lll|mmmm
   pha                                  // preserve token on stack

   and #$70                             // isolate literals count
   beq no_literals                      // skip if no literals to copy
   cmp #$70                             // literals_run_len?
   bne prepare_copy_literals            // if not, count is directly embedded in token

   jsr getsrc                           // get extra byte of variable literals count
                                        // the carry is always set by the cmp above
                                        // getsrc doesn't change it
   sbc #$f9                             // (literals_run_len)
   bcc prepare_copy_literals_direct
   beq large_varlen_literals            // if adding up to zero, go grab 16-bit count

   jsr getsrc                           // get single extended byte of variable literals count
   iny                                  // add 256 to literals count
   bcs prepare_copy_literals_direct     // (*like jmp prepare_copy_literals_direct but shorter)

large_varlen_literals:                  // handle 16 bits literals count
                                        // literals count = directly these 16 bits
   jsr getlargesrc                      // grab low 8 bits in x, high 8 bits in a
   tay                                  // put high 8 bits in y
   txa
   jmp prepare_copy_large_literals

prepare_copy_literals:
   tax
   lda shift_table-1,x                  // shift literals length into place
                                        //-1 because position 00 is reserved
prepare_copy_literals_direct:
   tax

prepare_copy_large_literals:
   beq copy_literals
   iny

copy_literals:

   jsr getput                           // copy one byte of literals
   dex
   bne copy_literals
   dey
   bne copy_literals
   
no_literals:
   pla                                  // retrieve token from stack
   pha                                  // preserve token again
   bmi get_long_offset                  // $80: 16 bit offset

   jsr getsrc                           // get 8 bit offset from stream in a
   tax                                  // save for later
   lda #$ff                             // high 8 bits
   bne got_offset                       // go prepare match
                                        // (*like jmp got_offset but shorter)

short_varlen_matchlen:
   jsr getsrc                           // get single extended byte of variable match len
   iny                                  // add 256 to match length

prepare_copy_match:
   tax
prepare_copy_match_y:
   txa
   beq copy_match_loop
   iny

copy_match_loop:
   lda $aaaa                            // get one byte of backreference
   jsr putdst                           // copy to destination



   // forward decompression -- put backreference bytes forward

   inc copy_match_loop+1
   beq getmatch_adj_hi
getmatch_done:



   dex
   bne copy_match_loop
   dey
   bne copy_match_loop
   beq decode_token                     // (*like jmp decode_token but shorter)



getmatch_adj_hi:
   inc copy_match_loop+2
   jmp getmatch_done


get_long_offset:                        // handle 16 bit offset:
   jsr getlargesrc                      // grab low 8 bits in x, high 8 bits in a

got_offset:



   //forward decompression - add match offset

   sta offshi+1                           // store high 8 bits of offset
   txa

   clc                                  // add dest + match offset
   adc putdst+1                         // low 8 bits
   sta copy_match_loop+1                // store back reference address
   
//offshi = off + 1

offshi:   lda #$aa                             // high 8 bits

   adc putdst+2
   sta copy_match_loop+2                // store high 8 bits of address
   


   pla                                  //retrieve token from stack again
   and #$0f                             //isolate match len (mmmm)
   adc #$02                             //plus carry which is always set by the high adc
   cmp #$12                             //match_run_len?
   bcc prepare_copy_match               //if not, count is directly embedded in token

   jsr getsrc                           //get extra byte of variable match length
                                        //the carry is always set by the cmp above
                                        //getsrc doesn't change it
   sbc #$ee                             //add match_run_len and min_match_size to match length
   bcc prepare_copy_match
   bne short_varlen_matchlen

                                        //handle 16 bits match length
   jsr getlargesrc                      //grab low 8 bits in x, high 8 bits in a
   tay                                  //put high 8 bits in y
                                        //large match length with zero high byte?
   bne prepare_copy_match_y             //if not, continue

decompression_done:
   rts

shift_table:
   .byte     $00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
   .byte $01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01
   .byte $02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02
   .byte $03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03
   .byte $04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04
   .byte $05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05
   .byte $06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06
   .byte $07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07



   //forward decompression -- get and put bytes forward

getput:
   jsr getsrc
putdst:
//lzsa_dst_lo = dst+1
//lzsa_dst_hi = dst+2
lzsa_dst:   sta $aaaa
   inc putdst+1
   beq putdst_adj_hi
   rts

putdst_adj_hi:
   inc putdst+2
   rts

getlargesrc:
   jsr getsrc                           //grab low 8 bits
   tax                                  //move to x
                                        //fall through grab high 8 bits

getsrc:
//lzsa_src_lo = src+1
//lzsa_src_hi = src+2
//src:   
lzsa_src: lda $aaaa
   inc getsrc+1
   beq getsrc_adj_hi
   rts

getsrc_adj_hi:
   inc getsrc+2
   rts

